/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package org.openzen.zenscript.javasource;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import org.openzen.zenscript.codemodel.FunctionParameter;
import org.openzen.zenscript.javashared.JavaClass;
import org.openzen.zenscript.javashared.JavaSynthesizedFunction;
import org.openzen.zenscript.javashared.JavaSynthesizedRange;
import org.openzen.zenscript.javashared.JavaSyntheticClassGenerator;

/**
 *
 * @author Hoofdgebruiker
 */
public class JavaSourceSyntheticTypeGenerator implements JavaSyntheticClassGenerator {
	private final File directory;
	private final JavaSourceFormattingSettings settings;
	private final JavaSourceContext context;
	
	public JavaSourceSyntheticTypeGenerator(File directory, JavaSourceFormattingSettings settings, JavaSourceContext context) {
		this.directory = new File(directory, "zsynthetic");
		this.settings = settings;
		this.context = context;
	}
	
	@Override
	public void synthesizeFunction(JavaSynthesizedFunction function) {
		JavaSourceImporter importer = new JavaSourceImporter(function.cls);
		JavaSourceTypeVisitor typeVisitor = new JavaSourceTypeVisitor(importer, context);
		
		StringBuilder contents = new StringBuilder();
		contents.append("@FunctionalInterface\n");
		contents.append("public interface ");
		contents.append(function.cls.getName());
		JavaSourceUtils.formatTypeParameters(typeVisitor, contents, function.typeParameters, false);
		contents.append(" {\n");
		contents.append(settings.indent);
		contents.append(function.header.getReturnType().accept(typeVisitor));
		contents.append(' ');
		contents.append("invoke(");
		boolean first = true;
		for (FunctionParameter parameter : function.header.parameters) {
			if (first) {
				first = false;
			} else {
				contents.append(", ");
			}
			contents.append(parameter.type.accept(typeVisitor));
			contents.append(' ');
			contents.append(parameter.name);
			if (parameter.variadic)
				contents.append("...");
		}
		contents.append(");\n");
		contents.append("}\n");
		
		writeFile(function.cls, importer, contents);
	}
	
	@Override
	public void synthesizeRange(JavaSynthesizedRange range) {
		JavaSourceImporter importer = new JavaSourceImporter(range.cls);
		JavaSourceTypeVisitor typeVisitor = new JavaSourceTypeVisitor(importer, context);
		
		StringBuilder contents = new StringBuilder();
		contents.append("public final class ").append(range.cls.getName()).append(" {\n");
		contents.append(settings.indent).append("public final ").append(range.baseType.accept(typeVisitor)).append(" from;\n");
		contents.append(settings.indent).append("public final ").append(range.baseType.accept(typeVisitor)).append(" to;\n");
		contents.append(settings.indent).append("\n");
		contents.append(settings.indent)
				.append("public ")
				.append(range.cls.getName())
				.append("(")
				.append(range.baseType.accept(typeVisitor))
				.append(" from, ")
				.append(range.baseType.accept(typeVisitor))
				.append(" to) {\n");
		contents.append(settings.indent).append(settings.indent).append("this.from = from;\n");
		contents.append(settings.indent).append(settings.indent).append("this.to = to;\n");
		contents.append(settings.indent).append("}\n");
		contents.append("}\n");
		
		writeFile(range.cls, importer, contents);
	}
	
	private void writeFile(JavaClass cls, JavaSourceImporter importer, StringBuilder contents) {
		if (!directory.exists())
			directory.mkdirs();
		
		File file = new File(directory, cls.getName() + ".java");
		try (Writer writer = new OutputStreamWriter(new BufferedOutputStream(new FileOutputStream(file)), StandardCharsets.UTF_8)) {
			writer.write("package zsynthetic;\n");
			
			JavaClass[] imports = importer.getUsedImports();
			if (imports.length > 0) {
				for (JavaClass import_ : imports) {
					if (import_.pkg.equals("java.lang"))
						continue;
					
					writer.write("import ");
					writer.write(import_.fullName);
					writer.write(";\n");
				}

				writer.write("\n");
			}
			
			writer.write('\n');
			writer.write(contents.toString());
		} catch (IOException ex) {
			ex.printStackTrace();
		}
	}
}
